/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jbpm.cal.xml;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.TimeZone;

import org.jbpm.cal.BusinessCalendar;
import org.jbpm.cal.Day;
import org.jbpm.cal.DayPart;
import org.jbpm.cal.Holiday;
import org.jbpm.xml.Parse;
import org.jbpm.xml.Parser;
import org.jbpm.xml.XmlUtil;
import org.w3c.dom.Element;

public class BusinessCalendarParser extends Parser {
  
  private static BusinessCalendarParser instance = null;
  
  public synchronized static BusinessCalendarParser getInstance() {
    if (instance==null) {
      instance = new BusinessCalendarParser();
    }
    return instance;
  }
  
  public static BusinessCalendar parseXmlString(String xmlString) {
    return (BusinessCalendar) getInstance()
            .createParse()
            .setString(xmlString)
            .execute()
            .checkProblems("business calendar string")
            .getDocumentObject();
  }

  public Object parseDocumentElement(Element element, Parse parse) {
    BusinessCalendar businessCalendar = new BusinessCalendar();
    
    TimeZone timeZone = null;
    if (element.hasAttribute("timezone")) {
      timeZone = TimeZone.getTimeZone(element.getAttribute("timezone"));
    } else {
      timeZone = TimeZone.getDefault();
    }
    businessCalendar.setTimeZone(timeZone);
    
    String hourFormatText = "HH:mm";
    if (element.hasAttribute("hour-format")) {
      hourFormatText = element.getAttribute("hour-format");
    }
    DateFormat hourFormat = new SimpleDateFormat(hourFormatText);

    Day[] days = new Day[8];
    Element daysElement = XmlUtil.element(element, "days");
    if (daysElement!=null) {
      days[Calendar.SUNDAY] = parseDay(daysElement, "sunday", hourFormat, businessCalendar, parse);
      days[Calendar.MONDAY] = parseDay(daysElement, "monday", hourFormat, businessCalendar, parse);
      days[Calendar.TUESDAY] = parseDay(daysElement, "tuesday", hourFormat, businessCalendar, parse);
      days[Calendar.WEDNESDAY] = parseDay(daysElement, "wednesday", hourFormat, businessCalendar, parse);
      days[Calendar.THURSDAY] = parseDay(daysElement, "thursday", hourFormat, businessCalendar, parse);
      days[Calendar.FRIDAY] = parseDay(daysElement, "friday", hourFormat, businessCalendar, parse);
      days[Calendar.SATURDAY] = parseDay(daysElement, "saturday", hourFormat, businessCalendar, parse);
    }
    businessCalendar.setDays(days);

    String dayFormatText = "dd/MM/yyyy";
    if (element.hasAttribute("day-format")) {
      dayFormatText = element.getAttribute("day-format");
    }
    DateFormat dayFormat = new SimpleDateFormat(dayFormatText);

    Holiday[] holidays = null;
    List<Element> holidayElements = XmlUtil.elements(element, "holiday");
    if (holidayElements!=null) {
      holidays = new Holiday[holidayElements.size()];
      for (int i=0; i<holidayElements.size(); i++) {
        holidays[i] = parseHoliday(holidayElements.get(i), dayFormat, businessCalendar, parse);
      }
    }
    businessCalendar.setHolidays(holidays);
    
    return businessCalendar;
  }

  private Day parseDay(Element daysElement, String dayText, DateFormat hourFormat, BusinessCalendar businessCalendar, Parse parse) {
    Day day = new Day();
    day.setBusinessCalendar(businessCalendar);

    Element dayElement = XmlUtil.element(daysElement, dayText);
    if (dayElement!=null) {
      DayPart[] dayParts = null;
      List<Element> partElements = XmlUtil.elements(dayElement, "part");
      if (partElements!=null) {
        dayParts = new DayPart[partElements.size()];
        for (int i=0; i<partElements.size(); i++) {
          dayParts[i] = parseDayPart(partElements.get(i), day, i, hourFormat, parse);
        }
      }
      day.setDayParts(dayParts);
    } else {
      parse.addProblem(dayText+" is not available in business calendar");
    }
    return day;
  }
  
  private DayPart parseDayPart(Element dayPartElement, Day day, int index, DateFormat hourFormat, Parse parse) {
    DayPart dayPart = new DayPart();
    dayPart.setDay(day);
    dayPart.setIndex(index);

    Calendar calendar = day.getBusinessCalendar().createCalendar();

    if (dayPartElement.hasAttribute("from")) {
      String fromText = dayPartElement.getAttribute("from");
      try {
        Date from = hourFormat.parse(fromText);
        calendar.setTime(from);
        dayPart.setFromHour(calendar.get(Calendar.HOUR_OF_DAY));
        dayPart.setFromMinute(calendar.get(Calendar.MINUTE));
      } catch (Exception e) {
        parse.addProblem("couldn't parse from attribute "+fromText);
      }
    } else {
      parse.addProblem("attribute from is required in part: "+XmlUtil.toString(dayPartElement));
    }

    if (dayPartElement.hasAttribute("to")) {
      String toText = dayPartElement.getAttribute("to");
      try {
        Date to = hourFormat.parse(toText);
        calendar.setTime(to);
        dayPart.setToHour(calendar.get(Calendar.HOUR_OF_DAY));
        dayPart.setToMinute(calendar.get(Calendar.MINUTE));
      } catch (Exception e) {
        parse.addProblem("couldn't parse from attribute "+toText);
      }
    } else {
      parse.addProblem("attribute from is required in part: "+XmlUtil.toString(dayPartElement));
    }

    return dayPart;
  }

  private Holiday parseHoliday(Element holidayElement, DateFormat dayFormat, BusinessCalendar businessCalendar, Parse parse) {
    Holiday holiday = new Holiday();
    try {
      if (holidayElement.hasAttribute("date")) {
        String holidayText = holidayElement.getAttribute("date");
        Date holidayDate = dayFormat.parse(holidayText.trim());
        holiday.setToDay(holidayDate);
        holiday.setFromDay(holidayDate);
        
      } else if ( (holidayElement.hasAttribute("from"))
                  && (holidayElement.hasAttribute("to"))
                ) {
        String dayText = holidayElement.getAttribute("from");
        Date dayDate = dayFormat.parse(dayText.trim());
        holiday.setFromDay(dayDate);

        dayText = holidayElement.getAttribute("to");
        dayDate = dayFormat.parse(dayText.trim());
        holiday.setToDay(dayDate);
      }

      // now we are going to set the toDay to the end of the day, rather then the beginning.
      // we take the start of the next day as the end of the toDay.
      Calendar calendar = businessCalendar.createCalendar();
      calendar.setTime(holiday.getToDay());
      calendar.add(Calendar.DATE, 1);
      Date toDay = calendar.getTime();
      holiday.setToDay(toDay);
      
    } catch (Exception e) {
      parse.addProblem("couldn't parse holiday: "+XmlUtil.toString(holidayElement));
    }

    return holiday;
  }
}
