/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jbpm.env.session;

import java.text.DateFormat;
import java.text.FieldPosition;
import java.text.SimpleDateFormat;
import java.util.Date;
import org.jbpm.log.Log;

import org.jbpm.PvmException;
import org.jbpm.cal.BusinessCalendar;
import org.jbpm.cal.Duration;
import org.jbpm.env.Environment;
import org.jbpm.env.context.JobContext;
import org.jbpm.jobexecutor.JobExecutor;
import org.jbpm.jobexecutor.JobSession;
import org.jbpm.pvm.Activity;
import org.jbpm.pvm.impl.ExecutionImpl;
import org.jbpm.pvm.impl.ObjectReference;

/**
 * @author Tom Baeyens
 * @author Pascal Verdage
 * @author Alejandro Guizar
 */
public class Timer extends Job<Boolean> {

  private static final long serialVersionUID = 1L;
  private static final Log log = Log.getLog(Timer.class.getName());

  static BusinessCalendar businessCalendar = new BusinessCalendar();

  protected String name;
  protected String repeat;
  protected ObjectReference<Activity> activityReference = null;
  // protected String transitionName = null;
  // protected ProcessElement processElement = null;

  public static final String EVENT_TIMER = "timer";

  protected Timer() {
    dueDate = new Date(Long.MAX_VALUE);
  }

  public Timer(ExecutionImpl execution) {
    super(execution);
    dueDate = new Date(Long.MAX_VALUE);
  }

  public Timer(String name, ExecutionImpl execution) {
    super(execution);
    if (name == null) {
      throw new PvmException("timer name is null");
    }
    this.name = name;
    dueDate = new Date(Long.MAX_VALUE);
  }

  public Boolean execute(Environment environment) throws Exception {
    log.debug("executing " + this);

    if (environment != null) {
      environment.addContext(new JobContext(this));
    }

    // TODO fire the event if there is a graph element specified

    if (activityReference != null) {
      try {
        activityReference.get().execute(execution);
      }
      catch (Exception activityException) {
        // NOTE that Errors are not caught because that might halt the JVM and
        // mask the original Error.
        log.info("WARNING: timer activity threw exception", activityException);

        // TODO if there is a processElement connected to this timer...
        // give that processElement a chance to catch the exception

        throw activityException;
      }
    }

    // TODO take a transition if one is specified

    boolean deleteThisJob = true;
    // if repeat is specified, reschedule the job
    if (repeat != null) {
      deleteThisJob = false;
      // suppose that it took the timer runner thread a very long time to execute the timers
      // then the repeat action dueDate could already have passed
      Duration duration = new Duration(repeat);
      Date nextDate = dueDate;
      do {
        nextDate = businessCalendar.add(nextDate, duration);
      } while (nextDate.getTime() <= System.currentTimeMillis());

      log.debug("rescheduled " + this + " for " + getDateFormat().format(nextDate));
      dueDate = nextDate;

      // release the lock on the timer
      setLockOwner(null);
      setLockExpirationTime(null);
      
      // notify the jobExecutor
      JobExecutor jobExecutor = Environment.getCurrent().get(JobExecutor.class);
      // the job executor is not present in the enterprise configuration, check before notifying 
      if (jobExecutor != null) {
        jobExecutor.jobWasAdded();
      }
    }
    else {
      // else delete the job
      log.debug("deleting " + this);
      environment.get(JobSession.class).delete(this);
    }

    return deleteThisJob;
  }

  @Override
  public String toString() {
    StringBuffer textRep = new StringBuffer();
    textRep.append("timer");
    if (name != null) {
      textRep.append('(').append(name);
      if (dueDate == null)
        textRep.append(')');
    }
    if (dueDate != null) {
      textRep.append(name != null ? ',' : '(');
      getDateFormat().format(dueDate, textRep, new FieldPosition(0));
      textRep.append(')');
    }
    return textRep.toString();
  }

  public static DateFormat getDateFormat() {
    return new SimpleDateFormat("yyyy-MM-dd HH:mm:ss,SSS");
  }

  public String getRepeat() {
    return repeat;
  }

  public void setRepeat(String repeat) {
    this.repeat = repeat;
  }

  public String getName() {
    return name;
  }

  public void setName(String name) {
    this.name = name;
  }

  public ObjectReference<Activity> getActivityReference() {
    return activityReference;
  }

  public void setActivityReference(ObjectReference<Activity> activityReference) {
    this.activityReference = activityReference;
  }
}
