/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jbpm.pvm;

import java.util.Date;
import java.util.List;
import java.util.Map;

/**
 * a graph (or tree) structure that can be executed.
 * 
 * <h2>Purpose</h2>
 * <p>ProcessDefinition is a base implementation that can be leveraged to build 
 * graph based execution languages.  While the ProcessDefinition class is concrete and 
 * can be used as-is (e.g. by aggregation), most likely processDefinition languages
 * will inherit from this ProcessDefinition and create more specialized implementations.  
 * </p>
 * 
 * <p>The specialized processDefinition language classes can extend this ProcessDefinition  
 * with new datastructures relevant for that perticular processDefinition language.
 * </p>
 * 
 * <h2>Structure</h2>
 * <p>A processDefinition contains a set of nodes.  Nodes can be connected with 
 * transitions or nodes can have nested nodes.  But the transitions and 
 * nested nodes can be combined.
 * </p>
 * 
 * <h2>Execution</h2>
 * <p>To create a new execution for a given processDefinition, see {@link #startExecution()}.
 * </p>
 * 
 * @author Tom Baeyens
 */
public interface ProcessDefinition extends CompositeElement {
  
  String EVENT_PROCESS_START = "process-start";
  String EVENT_PROCESS_END = "process-end";
  
  /** the package name of this process.  This is a similar way of 
   * grouping and creating a structure as in Java. We recommend 
   * the same conventions as with Java package names. */
  String getPackageName();

  /** automatically assigned during deployment of a process that 
   * represents the sequence number for process definitions with 
   * the same {@link ObservableElement#getName() name}. */ 
  int getVersion();

  /** start a new main path of execution for this process definition */ 
  Execution startExecution();

  /** start a new main path of execution for this process definition.
   * @param variables is a set of variables that is set as process variables 
   * before the process actually starts executing. */ 
  Execution startExecution(Map<String, Object> variables);

  /** start a new main path of execution for this process definition.
   * @param key is a user defined identifier for this new execution which 
   * has to be unique within the scope of this process definition. */ 
  Execution startExecution(String key);

  /** start a new main path of execution for this process definition. 
   * @param key is a user defined identifier for this new execution which 
   * has to be unique within the scope of this process definition.
  * @param variables is a set of variables that is set as process variables 
  * before the process actually starts executing. */ 
  Execution startExecution(String key, Map<String, Object> variables);

  /** the initial node that will be executed when a new process starts. */
  Node getInitial();

  /** variable declarations. */
  List<VariableDefinition> getVariableDefinitions();

  /** time this process definition got deployed. */
  Date getDeploymentTime();
}