/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jbpm.wire.binding;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.jbpm.wire.Descriptor;
import org.jbpm.wire.descriptor.MapDescriptor;
import org.jbpm.wire.xml.WireParser;
import org.jbpm.xml.Binding;
import org.jbpm.xml.Parse;
import org.jbpm.xml.Parser;
import org.jbpm.xml.XmlUtil;
import org.w3c.dom.Element;

/**
 * <p>This {@link Binding} specifies a {@link MapDescriptor}.</p>
 *
 * <p>A Map descriptor is defined by a <b><code>{@literal <map>}</code></b> xml element.</p>
 *
 * <p>This element can have an attribute "name", which specifies the name of the created object in the WireContext.</p>
 * <p>This element can have an attribute "class", which specifies the implementation of Map to use.
 * If this attribute is not present, the {@link HashMap} implementation is used.</p>
 *
 * <h4>Map entries</h4>
 * <p>Entries to add to the map can be specified with <b><code>{@literal <entry>}</code></b> elements.</p>
 * <p>Each &lt;entry&gt; must contain a <b><code>{@literal <key>}</code></b> element and a <b><code>{@literal <value>}</code></b> element.</p>
 * <p>Each &lt;key&gt; and &lt;value&gt; elements should contain a {@link Descriptor} element.</p>
 *
 * <h3>Example</h3>
 *
 * The following Xml declaration will create a TreeMap, and store it under the name 'm'.
 * The map will contain two elements:
 * <ul><li>'a' associated with the key 'foo'</li><li>'b' associated with the key 'bar'</li></ul>
 *
 * <pre> &lt;objects&gt;
 *   &lt;map name='m' class='java.util.TreeMap' &gt;
 *     &lt;entry&gt;
 *       &lt;key&gt; &lt;string value='foo'/&gt; &lt;/key&gt;
 *       &lt;value&gt; &lt;object name='a' class='java.lang.Object' /&gt; &lt;/value&gt;
 *     &lt;/entry&gt;
 *     &lt;entry&gt;
 *       &lt;key&gt; &lt;string value='bar'/&gt; &lt;/key&gt;
 *       &lt;value&gt; &lt;object name='b' class='java.lang.Object' /&gt; &lt;/value&gt;
 *     &lt;/entry&gt;
 *   &lt;/map&gt;
 * &lt;/objects&gt;</pre>
 *
 *
 * @author Tom Baeyens
 * @author Guillaume Porcher (documentation)
 *
 * @see WireParser
 */
public class MapBinding implements Binding {

  public Object parse(Element element, Parse parse, Parser parser) {
    MapDescriptor descriptor = new MapDescriptor();
    String className = XmlUtil.attribute(element,"class");
    if (AbstractCollectionBinding.verify(className, Map.class, parse, parser)) {
      descriptor.setClassName(className);
    }
    
    List<Descriptor> keyDescriptors = new ArrayList<Descriptor>();
    List<Descriptor> valueDescriptors = new ArrayList<Descriptor>();
    List<Element> elements = XmlUtil.elements(element);
    if (elements!=null) {
      for (Element entryElement: elements) {
        if ("entry".equals(XmlUtil.getTagName(entryElement))) {
          // key
          Element keyElement = XmlUtil.element(entryElement, "key");
          Element keyDescriptorElement = (keyElement!=null ? XmlUtil.element(keyElement) : null);
          Descriptor keyDescriptor = (Descriptor) parser.parseElement(keyDescriptorElement, parse, WireParser.CATEGORY_DESCRIPTOR);
          // value
          Element valueElement = XmlUtil.element(entryElement, "value");
          Element valueDescriptorElement = (valueElement!=null ? XmlUtil.element(valueElement) : null);
          Descriptor valueDescriptor = (Descriptor) parser.parseElement(valueDescriptorElement, parse, WireParser.CATEGORY_DESCRIPTOR);

          if ( (keyDescriptor!=null)
               && (valueDescriptor!=null)
             ) {
            keyDescriptors.add(keyDescriptor);
            valueDescriptors.add(valueDescriptor);
          } else {
            parse.addProblem("entry must have key and value element with a single descriptor as contents: "+XmlUtil.toString(entryElement));
          }
        } else {
          parse.addProblem("map can only contain entry elements: "+XmlUtil.toString(entryElement));
        }

      }
    }
    descriptor.setKeyDescriptors(keyDescriptors);
    descriptor.setValueDescriptors(valueDescriptors);
    return descriptor;
  }
  
  
}